#!/bin/sh
# \
exec tcl "$0" "$@"


#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Daggerfall CIF Access

#	(2)	Purpose:
#			To access the contents of the Daggerfall CIF-style
#			image collections.

#	(3)	Platform dependencies:
#			***  None  ***

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				Arena2Anime			--	Artwork main dialog

#			Exports:
#				browseTable			--	Report the directory of
#										iamges in a cif file
#				buildTables			--	Build the directory of
#										images in a cif file
#				cifFaces			--	Build the directory of
#										images in faces.cif
#				description			--	Answer the image description
#				imageAt				--	Answer the requested image
#										as a true color pixmap
#				shortAt				--	Answer the requested 16 bit
#										integer

#			Assertions about definitions:
#				Left-handed chirality is assumed.

#	(6)	References:

#			Library files:
#				Tcl					--	Base Tcl support
#				TclX				--	Extended Tcl support

#			Imports:
#				Arena2Config		--	Arena2 configuration
#				imgtoppm.tcl		--  True color converters

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********

#							 *****************
#							** Configuration **
#							 *****************

namespace eval Arena2Anime {
	variable GameDir		"/dos/g/game/dag/arena2"
	variable PalFile		"/dos/g/game/dag/arena2/art_pal.col"
	variable picFile		"pic,cif"
	variable ppmFile		"cif.00"
	variable xOffset		0
	variable yOffset		1
	variable width			2
	variable height			3
	variable compression	4
	variable imageSize		5
	variable headerStart	6
	variable frameCount		7
	variable frameSize		8

	variable tableHeaders	0
	variable tableFileName	1
	variable tableCount		2

	namespace export browseTable
	namespace export buildTables
	namespace export cifFaces
	namespace export description
	namespace export imageAt
	namespace export shortAt
}

if {[llength [namespace children :: ::Arena2Image]] == 0} then {
	source imgtoppm.tcl
	namespace import -force Arena2Image::blankImage
	namespace import -force Arena2Image::buildPalette
	namespace import -force Arena2Image::buildPaletteFrom
	namespace import -force Arena2Image::convertImage
	namespace import -force Arena2Image::convertRaw
	Arena2Image::buildPalette
}

#							 *****************
#							** Class methods **
#							 *****************

proc Arena2Anime::browseTable { cif } {
#		Browse the texture file with the specified serial number
#		and report its header structure.

	variable tableDesc
	variable tableCount
	variable tableHeaders

	set table [buildTables $cif]
	puts stdout "-- $cif [lindex $table $tableCount] images"
	puts stdout "-- Image header contents"
	reportImageHeaders [lindex $table $tableHeaders]
	return 0
}

proc Arena2Anime::buildTables { cif } {
#		Build the table of texture headers.

	variable tableHeaders

	set fid [Arena2Anime::open $cif]
	set table [headerTable $fid]
	lappend reply $table
	lappend reply "$cif"
	lappend reply [llength $table]
	Arena2Anime::close $fid
	return $reply
}

proc Arena2Anime::close { ttureFID } {
#		Close the texture stream.

	return [::close $ttureFID]
}

proc Arena2Anime::cifFaces { } {
#		Manufacture a header list for faces.cif.

	variable tableHeaders

	set ttureFID [open "faces.cif"]
	seek $ttureFID 0 end
	set largest [tell $ttureFID]
	close $ttureFID
	set index 0 
	while {$index < $largest} {
		set image [faceHeader]
		lappend image $index
		lappend imageHeaders $image
		incr index 4096
	}
	lappend reply $imageHeaders
	lappend reply "faces.cif"
	lappend reply [llength $imageHeaders]
	return $reply
}

proc Arena2Anime::description {text} {
#		Answer the image description.

	return "$text"
}

proc Arena2Anime::faceHeader {} {
#		Answer a header fabricated for faces.cif

#	lappend face 0 0 64 64 0 4096
	lappend face [binary format "s" 0]"
	lappend face [binary format "s" 0]"
	lappend face [binary format "s" 64]"
	lappend face [binary format "s" 64]"
	lappend face [binary format "s" 0]"
	lappend face [binary format "s" 4084]"
	return $face
}

proc Arena2Anime::findImageAt { serialNo index } {
#		Find the index-th image in a texture file.

	set tables [buildTables $serialNo]
	system xv [imageAt $tables $index]
}

proc Arena2Anime::headerTable { ttureFID } {
#		Answer the Tcl table of texture header information.

	variable headerOffset
	variable imageSize
	variable width
	variable height
	variable compression

	seek $ttureFID 0 end
	set largest [tell $ttureFID]
	seek $ttureFID 0 start
	set index 0 
	while {$index < $largest} {
		seek $ttureFID $index start
		set image [imageHeaderAsTcl $ttureFID]
#		lappend image $index
		set size [shortAt $image $imageSize]
		set size1 [expr [shortAt $image $height] * [shortAt $image $width]]
		if {$size == $size1} {
			set index [tell $ttureFID]
			incr index $size
			lappend image 1
			lappend image $size
		} else {
			binary scan [read $ttureFID 64] s* offsets
			set frames [lsearch -exact $offsets 0]
			if {$frames < 0} then {
				set frames 32
			}
			lappend image $frames
			set result [lindex $offsets end]
			if {$result < 0} then {
				set result [expr 65536 + $result]
			}
			lappend image $result
			incr index $result
		}
		lappend imageHeaders $image
	}
	return $imageHeaders
}

proc Arena2Anime::imageAt { table indx {dest {}}} {
#		Answer the indx-th image from the texture table.

	return [imagesAt $table $indx $dest]
}

proc Arena2Anime::imagesAt { table indx {dest {}}} {
#		Answer the indx-th set of images from the texture table.

	variable GameDir
	variable frameCount
	variable frameSize
	variable headerStart
	variable height
	variable imageSize
	variable picFile
	variable ppmFile
	variable tableFileName
	variable tableHeaders
	variable width
	variable xOffset
	variable yOffset

	if {[llength $dest] < 1} then {
		set outFile $ppmFile
	} else {
		set outFile $dest
	}
#	Clear the relics from the last animation.

	set blob [glob -nocomplain [file rootname $outFile].*]
	if {[llength $blob] > 0} then {
		eval file delete $blob
	}
	set header [lindex [lindex $table $tableHeaders] $indx]
	set fid [open [lindex $table $tableFileName]]
	set pic [::open $picFile w+]
	fconfigure $pic -translation binary
	set start [lindex $header $headerStart]
	set frames [lindex $header $frameCount]
	if {$frames > 1} then {
		set ww [shortAt $header $xOffset]
		set hh [shortAt $header $yOffset]
		set chunkSize [lindex $header $frameSize]
	} else {
		set ww [shortAt $header $width]
		set hh [shortAt $header $height]
		set chunkSize [shortAt $header $imageSize]
		set chunkSize [expr $chunkSize + 12]
	}
	seek $fid $start start
	fcopy $fid $pic -size $chunkSize
	::close $pic
	close $fid
	if {[lindex $table $tableFileName] == "faces.cif"} then {
		Arena2Image::convertRaw $picFile $outFile 64 64
	} elseif {$frames > 1} then {
#puts stdout "exec txttoppm -p $GameDir/art_pal.col \
			 $ww $hh $chunkSize 0x10002 $frames $outFile $picFile >txttoppm.out 2>txttoppm.err"
		exec txttoppm -p $GameDir/art_pal.col \
			 $ww $hh $chunkSize 0x10002 $frames $outFile $picFile >txttoppm.out 2>txttoppm.err
		set index 0
		set root [file rootname $outFile]
		set outFile {}
		while { $index < $frames } {
			lappend outFile [format "%s.%02d" $root $index]
			incr index
		}
	} else {
		::Arena2Image::convertImage $picFile $outFile
	}
	return $outFile
}

proc Arena2Anime::imageHeaderAsTcl { ttureFID } {
#		Read the next image header from the present position
#		in the file and answer its structure as a TCL list.

	set start [tell $ttureFID]
#	xOffset
	lappend reply [read $ttureFID 2]
#	yOffset
	lappend reply [read $ttureFID 2]
#	width
	lappend reply [read $ttureFID 2]
#	height
	lappend reply [read $ttureFID 2]
#	compression
	lappend reply [read $ttureFID 2]
#	imageSize
	lappend reply [read $ttureFID 2]
	lappend reply "$start"
	return "$reply"
}

proc Arena2Anime::open { tturePath } {
#		Open the requested texture pathname and return its handle.

	variable GameDir

	set path [file join "$GameDir" "$tturePath"]
	set fid [::open $path r]
	fconfigure $fid -translation binary
	return $fid
}

proc Arena2Anime::reportImageHeader { header } {
#		Document the contents of a texture image header.

	variable xOffset
	variable yOffset
	variable width
	variable height
	variable imageOffset
	variable imageSize
	variable headerStart
	variable compression
	variable frameCount

	set xOff [shortAt $header $xOffset]
	set yOff [shortAt $header $yOffset]
	set w [shortAt $header $width]
	set h [shortAt $header $height]
	set size [shortAt $header $imageSize]
	set frames [lindex $header $frameCount]
	set start [lindex $header $headerStart]
	puts stdout [format "(%d,%d) %dx%d (%d, %d) at %#x" \
		$xOff $yOff $w $h $size $frames $start]
	return 1
}

proc Arena2Anime::reportImageHeaders { headers } {
#		Report the contents of the current texture file image headers.

	set index 0
	set length [llength $headers]
	while {$index < $length} {
		set header [lindex $headers $index]
		puts -nonewline stdout "-- ($index): "
		reportImageHeader "$header"
		incr index
	}
	return 1
}

proc Arena2Anime::shortAt { aList index } {
#		Answer the integer value at the element-th entry of
#		a list and the index-th entry of that element.

	binary scan [lindex $aList $index] s result
	if {$result < 0} then {
		return [expr 65536 + $result]
	}
	return $result
}

#							 ******************
#							** Initialization **
#							 ******************

#	Inherit the installation configuration from the Arena2Config
#	namespace when it is available.

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Anime::GameDir "$Arena2Config::artDir"
	variable Arena2Anime::picFile [file join "$Arena2Config::tmpDir" "pic.cif"]
	variable Arena2Anime::ppmFile [file join "$Arena2Config::tmpDir" "ppm.00"]
	variable Arena2Anime::PalFile [file join "$Arena2Config::artDir" "art_pal.col"]
}
#Arena2Anime::browseTable 432
eval $argv
