#!/bin/sh
# \
exec wishx "$0" "$@"

#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Daggerfall Faction Viewer

#	(2)	Purpose:
#			To layout the Daggerfall faction organization as
#			a hierarchal list and animate selected contents.

#	(3)	Platform dependencies:
#			***  None  ***

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				Arena2Faction		--  Faction viewer class

#			Exports:
#				***  None  ***

#			Assertions about definitions:
#				***  None  ***

#	(6)	References:

#			Library files:
#				Tcl					--	Base Tcl support
#				TclX				--	Extended Tcl support
#				Tk					--	Base widget support

#			Imports:
#				Arena2Config		--	Arena2 configuration
#				cifimg.tcl			--	CIF art accessor
#				items.tcl			--	Texture are accessor

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********

#							 *****************
#							** Configuration **
#							 *****************

source cifimg.tcl
	namespace import -force Arena2Cif::imageAt
	namespace import -force Arena2Cif::shortAt
	namespace import -force Arena2Cif::buildTables
	namespace import -force Arena2Cif::description
	namespace import -force Arena2Cif::cifFaces
	namespace import -force Arena2Image::convertRaw

source items.tcl
	namespace import -force Arena2Items::browseItems
	namespace import -force Arena2Items::buildTables
	namespace import -force Arena2Items::picToTxt
	namespace import -force Arena2Items::txtImageAt

	namespace import -force Arena2Texture::animate
	namespace import -force Arena2Texture::clearAnimation

namespace eval Arena2Faction {
	variable factPath		"/dos/g/game/dag/arena2/faction.txt"
	variable flatsCfg		"/dos/g/game/dag/arena2/flats.cfg"
	variable tfacPath		"/dos/g/game/dag/arena2/tfac00i0.rci"
	variable facesPath		"/dos/g/game/dag/arena2/faces.cif"
	variable factTcl		"facts.tcl"
	variable flatTcl		"flats.tcl"
	variable face0Ppm		face0.ppm
	variable face1Ppm		face1.ppm
	variable flat0Ppm		flat1.00
	variable flat1Ppm		flat2.00
	variable hiddenFacts
	variable browsedFacts
	variable factions
	variable flats
	variable faces
}

#							 ********************
#							** Instance methods **
#							 ********************

proc Arena2Faction::addFactionAt { aList index childOf } {
#		Build the list of the presently visible factions from the source
#		faction list in aList.  Answer the list of visible faction names
#		in browsedFacts with descendent faction names properly indented
#		according to their respective positions in the faction hierarchy.

	variable hiddenFacts
	variable browsedFacts

	set size [llength $aList]
	while {$index < $size}  {
		set faction [lindex $aList $index]
		set findex $index
		incr index
		keylget faction "childOf" parent
		if { $parent != $childOf && $childOf > -1} then {
			return $index
		}
		keylget faction "name" name
		keylget faction "faction" serial
		keylget faction "indent" spaces
		if { [lsearch -exact $hiddenFacts $findex] < 0 } then {
			if { [lsearch -exact $hiddenFacts $index] > -1 } then {
				set child [lindex $aList $index]
				if { [keylget child "childOf"] == $serial } then {
					lappend browsedFacts $findex [replicate " " $spaces]$name...
				} else {
					lappend browsedFacts $findex [replicate " " $spaces]$name
				}
			} else {
				lappend browsedFacts $findex [replicate " " $spaces]$name
			}
		}
	}
	return $index
}

proc Arena2Faction::animators {} {
#		Answer the set of current animators.
	set reply {}
	if {[llength [namespace which -variable flat0]] > 0} then {
		variable flat0
		lappend reply $flat0
	}
	if {[llength [namespace which -variable flat1]] > 0} then {
		variable flat1
		lappend reply $flat1
	}
	return "$reply"
}

proc Arena2Faction::contents { w } {
#		Fill the contents of a listbox window with the hierarchal faction list.

	variable browsedFacts
	variable factions

	unset browsedFacts
	addFactionAt $factions 0 -1
    $w delete 0 end
    set each 1
    set final [llength $browsedFacts]
    while {$each < $final} {
    	$w insert end "[lindex $browsedFacts $each]"
    	incr each 2
    }
}

proc Arena2Faction::faceImageAt { index dest } {
#		Answer the index-th faces.cif entry in dest.

	variable faces
	set file [Arena2Cif::imageAt $faces $index $dest]
#	file copy -force $file $dest
}

proc Arena2Faction::factionAsFlat { index } {
#		Answer the face info for a faction index from the flat configuration.

	variable factions
	variable flats
	set faction [lindex $factions $index]
	keylget faction "face" face
	keylget faction "flat0" flat0
	set haveFlat1 [keylget faction "flat1" flat1]
	lappend reply [concat "face" $face]

#	Cross-reference this faction with flats.cfg to identify the
#	desired face photograph.

	set face0 [lsearch -regexp $flats [concat $flat0]]
	if {$face0 == -1} {
		puts "-- Can't locate face flat for flat0 $flat0"
#		puts "-- $faction"
	} else {
		set faction1 [lindex $flats $face0]
		keylget faction1 "face" face0
		if {$face0 != 410} {
			lappend reply [concat "face0" [list $face0]]
		}
		lappend reply [concat "flat0" [list $flat0]]
	}
	if { $haveFlat1 } then {

#		Cross-reference this faction with flats.cfg to identify the
#		desired face photograph.

		set face1 [lsearch -regexp $flats [concat $flat1]]
		if {$face1 == -1} {
			puts "-- Can't locate face flat for flat1 $flat1"
			puts "-- $faction"
			return $reply
		}
		set faction2 [lindex $flats $face1]
		keylget faction2 "face" face1
		if {$face1 != 410} {
			lappend reply [concat "face1" [list $face1]]
		}
		lappend reply [concat "flat1" [list $flat1]]
	} else {
#		lappend reply {face1 -1}
#		lappend reply {flat1 -1}
	}
	return $reply
}

proc Arena2Faction::factionAsTcl { } {
#		Read the converted faction.txt information as a TCL list.

	variable factTcl

	set fid [::open $factTcl r]
	lgets $fid facts
	::close $fid
	set facts [lindex $facts 0]
	return $facts
}

proc Arena2Faction::factionTable { } {
#		Transform the faction set in faction.txt into a sorted Tcl list.
#		The default sort is in child-descendant order.

	variable factTcl
	variable factPath

	if { [file exists $factTcl] } then {
		return [factionAsTcl]
	}
	system faction $factPath $factTcl
	return [factionAsTcl]
}

proc Arena2Faction::flatsAsTcl { } {
#		Read the converted faction.txt information as a TCL list.

	variable flatTcl

	set fid [::open $flatTcl r]
	lgets $fid flats
	::close $fid
	set flats [lindex $flats 0]
	return $flats
}

proc Arena2Faction::flatTable { } {
#		Transform the flats configuration in flats.cfg  into a Tcl list
#		of flat serial numbers and face codes.

	variable flatTcl
	variable flatsCfg

	if { [file exists $flatTcl] } then {
		return [flatsAsTcl]
	}
	system fromdos <$flatsCfg | awk -f flats.awk > $flatTcl
	return [flatsAsTcl]
}

proc Arena2Faction::initHidden { aList } {
#		Hide all the children of the principle factions
#		to begin with.

	variable hiddenFacts
	variable browsedFacts

	set size [llength $aList]
	set index 0
	while {$index < $size}  {
		set faction [lindex $aList $index]
		keylget faction "name" name
		keylget faction "childOf" parent
		if { $parent != -1 } then {
			lappend hiddenFacts $index
		}
		incr index
	}
	set browsedFacts {}
}

proc Arena2Faction::initialize { } {
#		Initialize the contents of the receiver.

	variable factions [factionTable]
	variable flats [flatTable]
	variable faces [Arena2Cif::cifFaces]
	initHidden $factions
}

proc Arena2Faction::popen { cmd } {
#		Emulate the popen function to catch the command output
#		from a cmd and return that result.

	flush stdout
	pipe fRead fWrite
	set newPid [fork]
	if {$newPid != 0} {
#		This is the parent thread.  Collect the output lines from
#		the child until it dies.

		set reply ""
		set fin 0
		signal error { SIGCHLD SIGPIPE }
		while { !$fin } {
			set fin [catch {gets $fRead line}]
			lappend reply $line
		}
		signal default { SIGCHLD SIGPIPE }
		close $fRead
		close $fWrite
		return $reply
	} else {
#		This is the child thread.

		dup $fWrite stdout
		lappend args -c "$cmd"
		execl /bin/sh $args
		kill [id process]
	}
}

proc Arena2Faction::tfacImageAt { index dest } {
#		Extract the tfac image for the index-th entry.

	variable tfacPath

	Arena2Image::convertRaw $tfacPath $dest 64 64 [expr $index * 4096]
}

proc Arena2Faction::txtImageAt { resNo index dest } {
#		Answer the index-th texture from texture.resNo in dest.

	variable Arena2Faction::faces
	set file [Arena2Cif::imageAt $faces $index]
	file copy -force $file $dest
}

proc positionWindow w {
#		Position a new window on the screen.
    wm geometry $w +300+300
}

proc loadFacts {w} {
#		Load a directory listbox with file names.

	Arena2Faction::contents $w.f.list
}

proc showFaction {w x y } {
#		Show the portrait(s) for the selected faction.

	variable Arena2Faction::browsedFacts
	variable Arena2Faction::face0Ppm
	variable Arena2Faction::face1Ppm
	variable Arena2Faction::flat0Ppm
	variable Arena2Faction::flat1Ppm

	set findex [lindex $browsedFacts [expr [$w.f.list index @$x,$y] << 1]]
	set portraits [Arena2Faction::factionAsFlat $findex]
puts "-- $portraits"
	keylget portraits "face" face
	set haveFace0 [keylget portraits "face0" face0]
	set haveFace1 [keylget portraits "face1" face1]
	set haveFlat0 [keylget portraits "flat0" flat0]
	set haveFlat1 [keylget portraits "flat1" flat1]
	face2 blank
	set faceImage1 face0
	set faceImage2 face1
	clearAnimation [Arena2Faction::animators]
	if { $haveFlat0 } {
		face0 blank
		if { $face != -1 } {
#			Display faces.cif face as face0.

			Arena2Faction::faceImageAt $face $face0Ppm
			face0 configure -file $face0Ppm
			set faceImage1 face1
			set faceImage2 face2
			if {$haveFace0 && $face0 == 0} then {
				set face0 -1
			}
		}
		if { $haveFace0 } {
			if { $face0 != -1 && $face0 != $face} {
#				Display tfac face0 as face0.

				Arena2Faction::tfacImageAt $face0 $face0Ppm
				$faceImage1 configure -file $face0Ppm
			}
		} else {
			$faceImage1 blank
			set face0 -1
		}
		animate ::Arena2Faction flat0 [lindex [Arena2Items::txtImageAt [concat $flat0] $flat0Ppm] 1]
#		flat0 configure -file $flat0Ppm
	} else {
		face0 blank
		flat0 blank
		set face0 -1
		set flat0 -1
	}
	if { !$haveFace1 } {
		set face1 $face0
	}
	if { $haveFlat1} {
#		Display face1 as face1 and flat1 as flat1.
		if { $face1 != $face0} {
			Arena2Faction::tfacImageAt $face1 $face1Ppm
			$faceImage2 configure -file $face1Ppm
		} else {
			$faceImage2 blank
		}
		if { $flat1 != $flat0 } {
			animate ::Arena2Faction flat1 [lindex [Arena2Items::txtImageAt [concat $flat1] $flat1Ppm] 1]
#			flat1 configure -file $flat1Ppm
		} else {
			flat1 blank
		}
	} else {
		face1 blank
		flat1 blank
	}
}

proc toggleHiddenFactions { w x y } {
#		Reveal or hide the descendents of the faction selected in
#		the listbox.

	variable Arena2Faction::browsedFacts
	variable Arena2Faction::hiddenFacts
	variable Arena2Faction::factions

	set sindex [$w.f.list index @$x,$y]
	set findex [lindex $browsedFacts [expr $sindex << 1]]
	set hindex [expr $findex + 1]
	set parent [lindex $factions $findex]
	set hidden [lsearch -exact $hiddenFacts $hindex]
	set pFact [keylget parent "faction"]
	set child [lindex $factions $hindex]
	set size [llength $factions]
	set needUpdate 0
	if { $hidden > -1 } then {

#		The first child of the selection is still hidden.  Reveal it
#		(and its siblings) by dropping them from the hidden set.

		while { $hindex < $size } {

#			If this entry is a child of the selection drop it from
#			the hidden set.

			if { [keylget child "childOf"] == $pFact } then {
				set hiddenFacts [lreplace $hiddenFacts $hidden $hidden]
				set needUpdate 1
			}
#			Advance to the next faction.

			incr hindex
			set hidden [lsearch -exact $hiddenFacts $hindex]
			set child [lindex $factions $hindex]
		}
	} else {

#		Check whether the selection has descendents which should be now
#		hidden from view.

		keylget parent "indent" indent
		while { $hindex < $size } {

#			If this entry is a child of the selection drop it from
#			the hidden set.

			if { [keylget child "indent"] > $indent } then {
				lappend hiddenFacts $hindex
				set needUpdate 1
			} else {
				break
			}
#			Advance to the next faction.

			incr hindex
			set hidden [lsearch -exact $hiddenFacts $hindex]
			set child [lindex $factions $hindex]
		}
	}
	if { $needUpdate } then {

#		Refresh the list box contents and restore the selection, if
#		modifications have occurred.

		loadFacts $w
		$w.f.list selection set $sindex
		$w.f.list see $sindex
	}
}

proc fcloseDialog { w } {
	clearAnimation [Arena2Faction::animators]
	set blob [glob -nocomplain \
		$Arena2Texture::picFile \
		$Arena2Faction::face0Ppm \
		$Arena2Faction::face1Ppm \
		[file rootname $Arena2Faction::flat0Ppm].?? \
		[file rootname $Arena2Faction::flat1Ppm].??]
	if {[llength $blob]} then {
		eval file delete $blob
	}
	catch {destroy $w}
}

proc fbuildDialog { } {
	variable Arena2Config::gamma
	set w .faction1
	catch {destroy $w}
	toplevel $w
	wm title $w "Daggerfall Factions"
	wm iconname $w "Image1"
	positionWindow $w

	label	$w.msg \
			-font {Helvetica 14} \
			-wraplength 4i \
			-justify left \
			-text "Select a faction"
	pack	$w.msg -side top

	frame	$w.buttons
	pack	$w.buttons \
			-side bottom \
			-fill x \
			-pady 2m
	button	$w.buttons.dismiss \
			-text Close \
			-command "fcloseDialog $w"
	pack	$w.buttons.dismiss \
			-side left \
			-expand 1

	frame	$w.spacer1 \
			-height 3m \
			-width 20
	label	$w.fileLabel \
			-text "Factions:"
	pack	$w.spacer1 $w.fileLabel \
			-side top \
			-anchor w
	frame	$w.f
	pack	$w.f \
			-side top \
			-expand yes \
			-fill both

	listbox	$w.f.list \
			-width 20 \
			-height 10 \
			-yscrollcommand "$w.f.scroll set"
	scrollbar $w.f.scroll \
			-command "$w.f.list yview"
	pack	$w.f.scroll \
			-side right \
			-fill y
	pack	$w.f.list \
			-side left \
			-fill both \
			-expand 1

	catch {image delete face0}
	image	create photo face0 \
			-gamma $Arena2Config::gamma
	catch {image delete flat0}
	image	create photo flat0 \
			-gamma $Arena2Config::gamma
	catch {image delete face1}
	image	create photo face1 \
			-gamma $Arena2Config::gamma
	catch {image delete flat1}
	image	create photo flat1 \
			-gamma $Arena2Config::gamma
	catch {image delete face2}
	image	create photo face2 \
			-gamma $Arena2Config::gamma

	frame	$w.right
	pack	$w.right \
			-side left \
			-expand yes \
			-padx 10 \
			-pady 10 \
			-fill both
	label	$w.right.face0 \
			-image face0 \
			-justify left
	label	$w.right.face1 \
			-image face1 \
			-justify left
	label	$w.right.flat0 \
			-image flat0 \
			-justify left
	label	$w.right.flat1 \
			-image flat1 \
			-justify left
	label	$w.right.face2 \
			-image face2 \
			-justify left

	grid	$w.right.flat0 \
			-column 0 -row 0 -rowspan 2
	grid	$w.right.face0 \
			-column 1 -row 0
	grid	$w.right.flat1 \
			-column 2 -row 0 -rowspan 2
	grid	$w.right.face1 \
			-column 1 -row 1
	grid	$w.right.face2 \
			-column 1 -row 2

	bind $w.f.list <Double-1> "showFaction $w %x %y"
	bind $w.f.list <Button-2> "toggleHiddenFactions $w %x %y"
#	bind $w <Button> "+$w.f.list selection set active"

	loadFacts $w
}
#							 ******************
#							** Initialization **
#							 ******************

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Faction::factPath	[file join "$Arena2Config::artDir" "faction.txt" ]
	variable Arena2Faction::flatsCfg	[file join "$Arena2Config::artDir" "flats.cfg" ]
	variable Arena2Faction::tfacPath	[file join "$Arena2Config::artDir" "tfac00i0.rci" ]
	variable Arena2Faction::facesPath	[file join "$Arena2Config::artDir" "faces.cif" ]
}
Arena2Faction::initialize
fbuildDialog
