#!/bin/sh
# the next line restarts using wish \
exec tcl "$0" "$@"


#								 ***********
#								** Summary **
#								 ***********

#						  Copyright 2001, Nagual Enterprises
#					Please see COPYING for the complete GNU license

#	(1)	Title:
#				Image to Portable Pixmap Converter

#	(2)	Purpose:
#			To convert Daggerfall pseudocolor images to true color
#			pixmaps.

#	(3)	Platform dependencies:
#			Left-handed chirality is assumed.

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				Arena2Image			--	Image converter class

#			Exports:
#				blankImage			--	Write a blank pixmap on a file
#				buildPalette		--	Import a Daggerfall palette file
#				buildPaletteFrom	--	Import a Daggerfall palette file
#				convertImage		--	Convert a Daggerfall image
#				convertRaw			--	Convert a raw Daggerfall pseudocolor
#										pixmap

#			Assertions about definitions:
#				An Arena2 palette must be imported before attempting
#				any image conversions.  Only one palette is available
#				at a time, but Arena2Image::currentPalette may be saved
#				to avoid rebuilding palettes over and over.

#	(6)	References:

#			Library files:
#				***  None  ***

#			Imports:
#				Arena2Config		--	Arena2 configuration settings

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********

namespace eval Arena2Image {
	variable exeDir				"/dos/g/game/dag"
	variable artDir				"/dos/g/game/dag/arena2"
	variable tmpDir				"/tmp"
	variable palette			"/dos/g/game/dag/arena2/art_pal.col"
	variable currentPalette		{}
	variable gamma				2.4
	namespace export blankImage
	namespace export buildPalette
	namespace export buildPaletteFrom
	namespace export convert
	namespace export convertImage
	namespace export convertRaw
}

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Config::artDir
	variable Arena2Config::exeDir
	variable Arena2Config::tmpDir
	variable Arena2Config::gamma
	variable Arena2Image::artDir "$Arena2Config::artDir"
	variable Arena2Image::tmpDir "$Arena2Config::tmpDir"
	variable Arena2Image::gamma "$Arena2Config::gamma"
	variable Arena2Image::palette [file join "$Arena2Config::artDir" "art_pal.col"]
}

proc Arena2Image::asPalette {colors} {
#		Answer the colors in a palette as a Tcl list.

	set pal {}
	foreach char $colors {
		binary scan $char c color
		lappend pal $color
	}
	return $pal
}

proc Arena2Image::blankImage {path width height} {
#		Create an empty width by height image on a path.

	set zip [binary format c3 {0 0 0}]
	set size [expr $width * $height]
	set out [::open $path w]
	fconfigure $out -translation binary
	writeRawPPMHeader $out $width $height 255
	while {$size > 0} {
		puts -nonewline $out "$zip"
		incr size -1
	}
	::close $out
	return 1
}

proc Arena2Image::buildPalette {} {
#		Build the stock palette as the default converter.

	variable palette
	set reply [buildPaletteFrom $palette]
	if { !$reply } {
		puts "Arena2Image: Build palette from $palette has failed"
	}
	return $reply
}

proc Arena2Image::buildPaletteFrom {palette {offset 0}} {
#		Build the palette index from the contents of a path.  Answer true
#		if the palette has been constructed, and false otherwise.

	variable currentPalette

	if { ![file exists $palette] } {
		puts "Arena2Image: cannot locate the image palette file '$palette'"
		return 0
	}
	set size [file size $palette]
	if {$size != 768 && $size != 776 && $offset == 0} {
		puts "Arena2Image: palette file '$palette' size not 768 or 772."
		return 0
	}
	set fid [::open $palette r]
	fconfigure $fid -translation binary
	if {$size == 776} {
		seek $fid 8 start
	}
	if {$offset > 0} then {
		seek $fid $offset start
	}
	set palBin [split [::read $fid 768] ""]
	::close $fid
	set currentPalette [asPalette $palBin]
	return 1
}

proc Arena2Image::convert {src dest {palFile {}}} {
#		Attempt to guess the format of the src image and
#		call the appropriate converted to record it on dest.

	variable currentPalette
	variable palette
	variable artDir
	if { ![file exists $src] } {
		puts "Arena2Image: source image $src does not exist."
		return 0
	}
	if {[llength $palFile]} then {
		variable palette [file join $artDir $palFile]
	}
	set size [file size $src]
	if {$size == 64000} then {
		return [convertRaw $src $dest 320 200 0]
	} elseif {$size == 68800} then {
		return [convertRaw $src $dest 320 215 0]
	} elseif { $size == 64768} then {
		set oldPal $currentPalette
		buildPaletteFrom $src 64000
		set reply [convertPalette $src $dest 320 200 0]
		set currentPalette $oldPal
		return $reply
	} else {
		return [convertImage $src $dest $size]
	}
	return 0
}

proc Arena2Image::convertImage {src dest {size 0}} {
#		Convert an Arena2 image from a src path to a ppm image
#		at a dest path.

	variable palette
	if { ![file exists $src] } {
		puts "Arena2Image: source image $src does not exist."
		return 0
	}
	set in [::open $src r]
	fconfigure $in -translation binary
	set header [readImage $in]
#	set header [lindex $headers 0]
	if {[lindex $header 0] == 0} {
		return 0
	}
	if {$size} then {
		set hSize [lindex $header 2]
		if {$hSize > $size || $hSize <= 0} then {
			return 0
		}
		if {[expr [lindex $header 0] * [lindex $header 1]] > $size} then {
			return 0
		}
	}
	if {[lindex $header 3]} then {
#		Pass this image off to the compression accellerator.

		variable tmpDir
		set pic [file join $tmpDir ppm.cvt]
		set out [::open $pic w]
		fconfigure $out -translation binary
		fcopy $in $out -size [lindex $header 2]
		::close $in
		::close $out
		exec txttoppm \
			-p $palette \
			[lindex $header 0] \
			[lindex $header 1] \
			[lindex $header 2] \
			0x02 \
			1 \
			$dest \
			$pic >txttoppm.out 2>txttoppm.err
		file delete $pic
	} else {
		::close $in
		exec txttoppm \
			-p $palette \
			[lindex $header 0] \
			[lindex $header 1] \
			[lindex $header 2] \
			1 \
			1 \
			$dest \
			$src
	}
	return 1
}

proc Arena2Image::convertPalette {src dest width height {offset 0}} {
#		Convert a raw Arena2 image from a src path to a ppm image
#		at a dest path.

	variable currentPalette

	if {![file exists $src]} {
		puts "Arena2Image: source image $src does not exist."
		return 0
	}
	set size [expr $width * $height]
	set out [::open raw.pal w]
	fconfigure $out -translation binary
	puts -nonewline $out [binary format c* $currentPalette]
	::close $out
	exec txttoppm \
			-p raw.pal \
			$width \
			$height \
			$size \
			0 \
			1 \
			$dest \
			$src
	file delete raw.pal
	return 1
}

proc Arena2Image::convertRaw {src dest width height {offset 0}} {
#		Convert a raw Arena2 image from a src path to a ppm image
#		at a dest path.

	variable palette
	if {![file exists $src]} {
		puts "Arena2Image: source image $src does not exist."
		return 0
	}
	set size [expr $width * $height]
	if {$offset} then {
		set subs -$offset
	} else {
		set subs 1
	}
	exec txttoppm \
			-p $palette \
			$width \
			$height \
			$size \
			0 \
			$subs \
			$dest \
			$src
	return 1
}

proc Arena2Image::readImage {fid} {
#		Read a full Arena2 image from a file strream.  Answer the image
#		bytes or nil on error.

	set header [readRawHeader $fid]
	if {[llength $header] == 0} {
		return $header
	}
	if {[lindex $header 0] == 0} {
		return $header
	}
	if {[lindex $header 3]} then {
#		Compressed image; don't need data stream here.

		return $header
	}
	return $header
#	return [list $header [readRaw $fid [lindex $header 2]]]
}

proc Arena2Image::readRawHeader {fid} {
#		Read the image header from a stream and answer a three-element
#		list {width height imageBytes}

	set header [read $fid 12]
	binary scan $header ssssss canvasX canvasY imageWidth imageHeight zip1 imageSize
	if {$imageSize < 0} {
		set imageSize [expr 65536 + $imageSize]
	}
	if {$imageWidth < 0} {
		set imageWidth [expr 65536 + $imageWidth]
	}
	if {$imageHeight < 0} {
		set imageHeight [expr 65536 + $imageHeight]
	}
	lappend reply $imageWidth $imageHeight $imageSize
	set size [expr $imageHeight * $imageWidth]
	lappend reply [expr $size != $imageSize]
#	if {$size != $imageSize} {
#		puts "Arena2Image: image size isn't a produce of width * height."
#		puts "Arena2Image: $imageSize isn't $imageWidth * $imageHeight."
#		return {}
#	}
	return $reply
}

proc Arena2Image::readRaw { fid bytes } {
#		Read the next bytes from a stream and answer the result.

	return [read $fid $bytes]
}

proc Arena2Image::writePalette {palette} {
#		Build the palette index from the contents of a path.  Answer true
#		if the palette has been constructed, and false otherwise.

	variable currentPalette

	set fid [::open $palette w]
	fconfigure $fid -translation binary
	puts -nonewline $fid [binary format c* $currentPalette]
	::close $fid
	return 1
}

proc Arena2Image::writeRawPPMHeader {fid width height depth} {
#		Write the PPM raw image header.

	puts $fid "P6"
	puts $fid "$width $height $depth"
	return 1
}

proc Arena2Image::writeRawPPM {fid rawData count} {
#		Write count bytes of raw PPM data on a file stream.

	variable currentPalette
	set index 0
	set rawSize [expr $count * 3]
	binary scan $rawData c* chars
#	while {$index < $count} 
	foreach char $chars {
		set char [lindex $chars $index]
		if {$char < 0} {
			set char [expr 256 + $char]
		}
		set first [expr 3 * $char]
		set last [expr 2 + $first]
		append result [lrange $currentPalette $first $last] " "
		incr index
	}
	puts -nonewline $fid [binary format c* $result]
	return 1
}

#namespace import -force Arena2Image::convertRaw
#namespace import -force Arena2Image::buildPalette
#Arena2Image::buildPalette
#Arena2Image::convertRaw /dos/g/game/dag/save0/image.raw junk 80 40
#eval $argv
