#!/bin/sh
# \
exec tcl "$0" "$@"

#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Daggerfall Item Catalogue

#	(2)	Purpose:
#			To access the Daggerfall item catalogue stored in
#			fall.exe

#	(3)	Platform dependencies:
#			***  None  ***

#	(4)	Change history:
#			***  None  ***

#	(5)	Definitions:

#			Namespaces:
#				Arena2Items			--	Class to access the item
#										catalogue stored in fall.exe

#			Exports:
#				browseItems			--	Report the item catalogue
#										entries
#				buildTables			--	Build the list of catalogue
#										entries
#				picToTxt			--	Answer the texture#/image#
#										from a picture code
#				txtImageAt			--	Answer the texture(s) for
#										a requested item.

#			Assertions about definitions:
#				***  None  ***

#	(6)	References:

#			Library files:
#				Tcl					--	Base Tcl support

#			Imports:
#				texture.tcl			--	Support for item graphics

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********

#							 *****************
#							** Configuration **
#							 *****************

source texture.tcl

namespace eval Arena2Items {
	variable ExeDir			"/dos/g/game/dag"
	variable tmpDir			"/tmp"
	variable itemName		0
	variable itemPicture1	1
	variable itemPicture2	2
	variable gamma			2.4

	namespace export browseItems
	namespace export buildTables
	namespace export picToTxt
	namespace export txtImageAt
}

#							 ********************
#							** Instance methods **
#							 ********************

proc Arena2Items::browseItems { } {
#		Display a report of the known Daggerfall items
#		and their item locations.

	set items [buildTables]
	set size [llength $items]
	set item 0
	while { $item < $size } {
		reportItem [lindex $items $item]
		incr item
	}
}

proc Arena2Items::buildTables { } {
#		Build the list of known Daggerfall items from fall.exe

	set fid [open "fall.exe"]
	seek $fid 1796138 start
	set item 0
	while { $item < 288 } {
		set next [nextItem $fid]
		lappend reply $next
		incr item
	}
	close $fid
	return $reply
}

proc Arena2Items::byteAt { aList index } {
#		Answer the integer value at the element-th entry of
#		a list and the index-th entry of that element.

	binary scan [lindex $aList $index] c result
	return $result
}

proc Arena2Items::close { ttureFID } {
#		Close the fall.exe stream.

	return [::close $ttureFID]
}

proc Arena2Items::longAt { aList index } {
#		Answer the integer value at the element-th entry of
#		a list and the index-th entry of that element.

	binary scan [lindex $aList $index] i result
	return $result
}

proc Arena2Items::nextItem { fid } {
#		Answer the next item from fall.exe.

	binary scan [read $fid 24] A* name
	lappend reply $name
	read $fid 20
	binary scan [read $fid 2] s pic
	lappend reply $pic
	binary scan [read $fid 2] s pic
	lappend reply $pic
	return $reply
}

proc Arena2Items::open { tturePath } {
#		Open the requested image pathname and return its handle.

	variable ExeDir

	set path [file join "$ExeDir" "$tturePath"]
	set fid [::open $path r]
	fconfigure $fid -translation binary
	return $fid
}

proc Arena2Items::picToAscii { pic } {
#		Convert the given picture code to an ascii string.

	set texture [picToTxt $pic]
	return "[lindex $texture 0].[lindex $texture 1]"
}

proc Arena2Items::picToTxt { pic } {
#		Convert the given picture code into a texture file
#		resource number and entry within that resource pool.

	set resNo [expr $pic >> 7]
	if { $resNo < 0 } then {
		set delta [expr 256 + $resNo]
		set resNo [expr 256 + $delta]
	}
	lappend reply $resNo
	lappend reply [expr $pic & 127]
	return $reply
}

proc Arena2Items::reportItem { item } {
#		Describe an item on stdout.

	variable itemName
	variable itemPicture1
	variable itemPicture2

	set pic1 [picToAscii [lindex $item $itemPicture1]]
	set pic2 [picToAscii [lindex $item $itemPicture2]]
	puts "-- $pic1 $pic2	[lindex $item $itemName]"
	return
}

proc Arena2Items::shortAt { aList index } {
#		Answer the integer value at the element-th entry of
#		a list and the index-th entry of that element.

	binary scan [lindex $aList $index] s result
	return $result
}

proc Arena2Items::txtImageAt { resource dest } {
	set resNo [lindex $resource 0]
	set serialNo [lindex $resource 1]
	set texture [Arena2Texture::buildTables $resNo]
	lappend reply "$resNo: $serialNo"
	lappend reply [Arena2Texture::imagesAt $texture $serialNo $dest]
	return $reply
}

#							 ********************
#							** Instance methods **
#							 ********************

#	Inherit the installation configuration from the Arena2Artwork
#	namespace when it is available.

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Config::artDir
	variable Arena2Config::exeDir
	variable Arena2Config::tmpDir
	variable Arena2Config::gamma
	variable Arena2Items::ExeDir "$Arena2Config::exeDir"
	variable Arena2Items::tmpDir "$Arena2Config::tmpDir"
	variable Arena2Items::gamma "$Arena2Config::gamma"
}
