#!/bin/sh
# \
exec tcl "$0" "$@"

#								 ***********
#								** Summary **
#								 ***********

#						  Copyright 2001, Nagual Enterprises
#					Please see COPYING for the complete GNU license

#	(1)	Title:
#				Texture file 

#	(2)	Purpose:
#			To provide a high-level API for accessing the individual
#			picture elements contained in a Daggerfall texture file.

#	(3)	Platform dependencies:
#			Left-handed chirality is assumed by the data stream
#			converters.

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				Arena2Texture		--	Class accessing the contents
#										of the standard Daggerfall
#										texture files.

#			Exports:
#				animate				--	Add and activate a new animation
#										sequence to the application space
#				browseTable			--	Document the directory contents of
#										a texture file
#				buildTables			--	Answer the directory contents of a
#										texture file as a Tcl list
#				clearAnimation		--	Stop all the animations in the
#										application space
#				currentRate			--	Answer the latest animation rate
#										from the current texture file
#				currentRate:		--	Assign the animation rate for
#										the current texture file
#				description			--	Answer the descriptive comment
#										from a texture file
#				findImageAt			--	Display a texture image in an
#										external viewer
#				imageAt				--	Answer a single image frame
#										converted to a portable pixmap
#				imagesAt			--	Answer all the frames in one image
#										after conversion to portable pixmap
#				longAt				--	Answer the 32 bit entity as a Tcl
#										number
#				ppmFile				--	The (temporary) image pixmap
#				shortAt				--	Answer the 16 bit entity as a Tcl
#										number

#			Assertions about definitions:
#				The bit accessors assume left-handed chirality.

#				The findImageAt method assumes that `xv' has been installed.

#				Some texture files (000 and 001) aren't textures at all, but
#				a sort of color palette index.  There are no images inside,
#				only color mappings, 0 to 127 and 128 to 255.

#	(6)	References:

#			Library files:
#				Tcl					--	Base Tcl support
#				Tclx				--	Extended Tcl support
#				txttoppm			--	Accellerated image to pixmap
#										converter

#			Imports:
#				***  None  ***

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********


namespace eval Arena2Texture {
	variable GameDir		"/dos/g/game/dag/arena2"
	variable picFile		"/tmp/pic"
	variable ppmFile		"/tmp/pic.00"

#	Texture directory fields

	variable headerType1	0
	variable headerOffset	1
	variable headerType2	2
	variable headerMystery	3

#	Image header fields

	variable xOffset		0
	variable yOffset		1
	variable width			2
	variable height			3
	variable compression	4
	variable imageSize		5
	variable imageOffset	6
	variable interlace		7
	variable subimages		8
	variable frameDelay		9
	variable imageScale		10
	variable headerStart	11

#	Texture table fields

	variable tableOffsets	0
	variable tableHeaders	1
	variable tableDesc		2
	variable tableFileName	3
	variable tableCount		4

#	Animation frame block fields

	variable frameSet		0
	variable currentFrame	1
	variable maxFrame		2
	variable frameWindow	3
	variable frameRate		4
	variable frameSpeed		5
	variable animator		6

	variable defaultRate	300
	variable _currentRate	0

	namespace export animate
	namespace export browseTable
	namespace export buildTables
	namespace export clearAnimation
	namespace export currentRate
	namespace export currentRate:
	namespace export description
	namespace export findImageAt
	namespace export imageAt
	namespace export imagesAt
	namespace export longAt
	namespace export ppmFile
	namespace export shortAt
}
#	Image compression codes:
#		0x0002		--	RLE encoding by whole image with 8-bit count fields
#		0x0003		--	RLE 8-bit count encoding up to 31 animation frames
#		0x1108		--	Image compressed row by row; imageSize is gold;
#		0x0108			RLE encoding uses 16-bit count fields
#		0x0900		--	Image interlaced on 256 x nnn canvas with
#		0x0100			other 0x0900 images.  Interlace field gives
#		0x0101			the interlace offset to skip to reach the
#						next row of the image.  imageSize does not reflect
#						the byte stream count; 256 * imageHeight is the
#						proper byte count to extract.

proc Arena2Texture::clearAnimation {ids} {
#		Stop all the animations in the application.

	foreach anime $ids {
		after cancel $anime
	}
}

proc Arena2Texture::animate { self window fileSet } {
#		Animate the flipbook of cells indexed by a fileSet in
#		the specified image window.

	variable frameSet
	variable currentFrame
	variable maxFrame
	variable frameWindow
	variable defaultRate
	variable _currentRate

#	Build the frame block for this animation:
#		the file set
#		the current frame number
#		the maximum frame number
#		the animation pane
#		the default frame rate (unused)
#		the frame rate from the texture file (in milliseconds)

	lappend anime $fileSet
	lappend anime 0
	lappend anime [llength $fileSet]
	lappend anime $window
	lappend anime $defaultRate

#	Use the animation rate extracted by the last texture pull
#	for this block.  NPCs posed for dialog interaction have
#	rates coded as 71, but this results in the infamous
#	`whirling dervish' effect.  Slow them down to a bus stop
#	fidget.  Empirically, 1185 is the largest (and thus slowest)
#	rate found, and corresponds to two frames per second.

#	The animator requires a millisecond interval, so oblige it.

	set speed [expr ($_currentRate * 2000) / 1185]
	if {$speed == 0 } then {
		set speed 10000
	}
	lappend anime $speed
	lappend anime $self
	animationStep $anime
}

proc Arena2Texture::animationStep { frameBlock } {
#		Advance the animation sequence to the next frame and display it.
#		Schedule the next display event.

	variable frameSet
	variable currentFrame
	variable maxFrame
	variable frameWindow
	variable frameRate
	variable frameSpeed
	variable animator

	set object [lindex $frameBlock $animator]
	set next [lindex $frameBlock $currentFrame]
	set index $next
	incr next
	if { $next >= [lindex $frameBlock $maxFrame] } then {
		set next 0
	}
	set frameBlock [lreplace $frameBlock $currentFrame $currentFrame $next]

#	Schedule the next frame of the animation.

	lappend joined Arena2Texture::animationStep
	lappend joined $frameBlock
	set id [after [lindex $frameBlock $frameSpeed] $joined]

#	Display the current animation frame in its window.

	set pane [lindex $frameBlock $frameWindow]
	$pane configure -file [lindex [lindex $frameBlock $frameSet] $index]
	namespace eval $object variable $pane $id
	return $id
}

proc Arena2Texture::browseAllTables {} {
#		Browse all the texture files in the game. */

	set count 0
	while {$count < 512 } {
		Arena2Texture::browseTable [format %03d $count]
		puts ""
		incr count
	}
}

proc Arena2Texture::browseTable { serialNo } {
#		Browse the texture file with the specified serial number
#		and report its header structure.

	variable tableDesc
	variable tableOffsets
	variable tableCount
	variable tableHeaders

	set table [buildTables $serialNo]
	if {[llength $table] == 0} then {
		return 1
	}
	puts stdout "-- texture.$serialNo [format %3d [lindex $table $tableCount]] images, [lindex $table $tableDesc]"
	puts stdout "-- Header contents:"
	reportHeaders [lindex $table $tableOffsets] $serialNo
	puts stdout ""
	puts stdout "-- Image header contents:"
	reportImageHeaders [lindex $table $tableHeaders] $serialNo
	return 0
}

proc Arena2Texture::buildTables {serialNo} {
#		Build the table of texture headers.

	set text "texture.$serialNo"
	set fid [Arena2Texture::open $text]
	if {$fid == -1} then {
		return {}
	}
	binary scan [type $fid] s size
	binary scan [describe $fid] A* desc
	if {[clength "$desc"] < 1} then {
		set desc "$text"
	}
	set table [headerTable $fid $size]
	if {[llength $table] == 0 } then {
		lappend table {}
	}
	lappend table "$desc"
	lappend table "$text"
	lappend table "$size"
	Arena2Texture::close $fid
	return $table
}

proc Arena2Texture::close { ttureFID } {
#		Close the texture stream.

	return [::close $ttureFID]
}

proc Arena2Texture::currentRate {} {
#		Answer the last animation frame rate encountered.

	variable _currentRate
	return $_currentRate
}

proc Arena2Texture::currentRate: {rate} {
#		Assign the last animation frame rate encountered.

	variable _currentRate
	set _currentRate $rate
}

proc Arena2Texture::describe { ttureFID } {
#		Answer the description tag from the texture file

	seek $ttureFID 2 start
	return [read $ttureFID 24]
}

proc Arena2Texture::description {text} {
#		Build the table of texture headers.

	set fid [Arena2Texture::open $text]
	binary scan [type $fid] s type
	binary scan [describe $fid] A* desc
	Arena2Texture::close $fid
	if {[clength "$desc"] < 1} then {
		set desc "$text"
	}
	return "$desc"
}

proc Arena2Texture::findImageAt { serialNo index } {
#		Find the index-th image in a texture file.

	set tables [buildTables $serialNo]
	system xv [imageAt $tables $index]
}

proc Arena2Texture::headerAsTcl { ttureFID } {
#		Read the next texture header from the present position
#		in the file and answer its structure as a TCL list.

	set type1 [read $ttureFID 2]
	set offset [read $ttureFID 4]
	set type2 [read $ttureFID 4]
	set mystery [read $ttureFID 2]
	read $ttureFID 8
	lappend reply $type1
	lappend reply $offset
	lappend reply $type2
	lappend reply $mystery
	return "$reply"
}

proc Arena2Texture::headersAsTcl { ttureFID tableSize } {
#		Answer the set of texture headers as an unsorted TCL list.

	seek $ttureFID 0 end
	set smallest [tell $ttureFID]
	seek $ttureFID 26 start
#	set current [tell $ttureFID]
	set count $tableSize

#	Iterate through the the image group headers.

	while {$count > 0} {
		set header [headerAsTcl $ttureFID]
#		set offset [longAt $header 1]
#		if {$offset < $smallest} {
#			set smallest $offset
#		}
#		set current [tell $ttureFID]
		lappend headers "$header"
		incr count -1
	}
	return "$headers"
}

proc Arena2Texture::headerTable { ttureFID tableSize } {
#		Answer the Tcl table of texture header information.

	variable headerOffset
	variable height

	seek $ttureFID 0 end
	set eof [tell $ttureFID]
	set headers [headersAsTcl $ttureFID $tableSize]
	set entryCount [llength $headers]
	set index 0
	while {$index < $entryCount} {
		set off [longAt [lindex $headers $index] $headerOffset]
		if {$off == 0 || $off == $eof } then {
			lappend reply "$headers" {}
			return "$reply"
		}
		seek $ttureFID $off start
		set image [imageHeaderAsTcl $ttureFID]
		lappend image $off
		lappend imageHeaders $image
		incr index
	}
	lappend reply "$headers" "$imageHeaders"
	return "$reply"
}

proc Arena2Texture::imageAt { table indx } {
#		Answer the indx-th image from the texture table.

	variable tableHeaders
	variable tableFileName
	variable imageSize
	variable imageOffset
	variable width
	variable height
	variable compression
	variable picFile
	variable ppmFile
	variable headerStart
	variable interlace
	variable subimages
	variable GameDir

	set header [lindex [lindex $table $tableHeaders] $indx]
	set chunkSize [longAt $header $imageSize]
	set ww [shortAt $header $width]
	set hh [shortAt $header $height]
	set compress [shortAt $header $compression]
	set fid [open [lindex $table $tableFileName]]
	set pic [::open $picFile w+]
	fconfigure $fid -translation binary
	fconfigure $pic -translation binary
	set start [lindex $header $headerStart]
	set offset [longAt $header $imageOffset]
	if { $offset == 0 } then {
		puts "Entry has no image data."
		Arena2Image::blankImage $ppmFile 16 16
		return $ppmFile
	}
	set offset [expr $start + $offset]
	set u2 [shortAt $header $interlace]
	set subs [shortAt $header $subimages]
	seek $fid $offset start
	if { $subs > 1 } then {
	} elseif { $compress == 2304 || $compress == 256 || $compress == 257 } then {
		set chunkSize [expr $hh * 256]
	}
	if { $compress == 0 && $u2 > 0 } then {
		set compress 2304
		set chunkSize [expr $hh * 256]
	}
	fcopy $fid $pic -size $chunkSize
	::close $pic
	close $fid
#	puts "system txttoppm -p $GameDir/art_pal.col $ww $hh $chunkSize $compress $subs $ppmFile $picFile"
	exec txttoppm -p $GameDir/art_pal.col $ww $hh $chunkSize $compress $subs $ppmFile $picFile
	return $ppmFile
}

proc Arena2Texture::imagesAt { table indx dest } {
#		Answer the indx-th image from the texture table.
#		A texture with multiple images (animation) will
#		create multiple image files (*.00, *.01, etc).

	variable tableHeaders
	variable tableFileName
	variable imageSize
	variable imageOffset
	variable width
	variable height
	variable compression
	variable picFile
	variable headerStart
	variable interlace
	variable subimages
	variable GameDir
	variable _currentRate
	variable frameDelay

	set header [lindex [lindex $table $tableHeaders] $indx]
	if {[llength $header] == 0} then {
		return {}
	}
#	Clear the relics from the last animation.

	set blob [glob -nocomplain [file rootname $dest].*]
	if {[llength $blob] > 0} then {
		eval file delete $blob
	}
	set _currentRate [shortAt $header $frameDelay]
	set chunkSize [longAt $header $imageSize]
	set ww [shortAt $header $width]
	set hh [shortAt $header $height]
	set compress [shortAt $header $compression]
	set fid [open [lindex $table $tableFileName]]
	set pic [::open $picFile w+]
	fconfigure $fid -translation binary
	fconfigure $pic -translation binary
	set start [lindex $header $headerStart]
	set offset [longAt $header $imageOffset]
	set offset [expr $start + $offset]
	set u2 [shortAt $header $interlace]
	set subs [shortAt $header $subimages]
	seek $fid $offset start
	if { $subs > 1 } then {
	} elseif { $compress == 2304 || $compress == 256 || $compress == 257 } then {
		set chunkSize [expr $hh * 256]
	}
	if { $compress == 0 && $u2 > 0 } then {
		set compress 2304
		set chunkSize [expr $hh * 256]
	}
	fcopy $fid $pic -size $chunkSize
	::close $pic
	close $fid

#	`system' doesn't seem to work under Wintel, so use `exec' and ignore the
#	output.  (`system' complains it cannot start the executable, which is
#	a fiction)

#	puts "exec txttoppm -p $GameDir/art_pal.col $ww $hh $chunkSize $compress $subs $dest $picFile >txttoppm.out 2>txttoppm.err"
	exec txttoppm -p $GameDir/art_pal.col $ww $hh $chunkSize $compress $subs $dest $picFile >txttoppm.out 2>txttoppm.err
	if { $subs < 1 } then {
		set subs  1
	}
	set index 0
	set root [file rootname $dest]
	while { $index < $subs } {
		lappend reply [format "%s.%02d" $root $index]
		incr index
	}
	return $reply
}

proc Arena2Texture::imageHeaderAsTcl { ttureFID } {
#		Read the next image texture header from the present position
#		in the file and answer its structure as a TCL list.

	set start [tell $ttureFID]
#	xOffset
	lappend reply [read $ttureFID 2]
#	yOffset
	lappend reply [read $ttureFID 2]
#	width
	lappend reply [read $ttureFID 2]
#	height
	lappend reply [read $ttureFID 2]
#	compression
	lappend reply [read $ttureFID 2]
#	imageSize
	lappend reply [read $ttureFID 4]
#	imageOffset
	lappend reply [read $ttureFID 4]
#	interlace
	lappend reply [read $ttureFID 2]
#	subimagesNo
	lappend reply [read $ttureFID 2]
#	frameDelay -- The animation frame delay number
	lappend reply [read $ttureFID 2]
#	imageScale -- Actually is 2 shorts
#	xScale, yScale to enlarge or reduce the appearance of the
#	image in the 3-D view.
	lappend reply [read $ttureFID 4]
	lappend reply "$start"
	return "$reply"
}

proc Arena2Texture::longAt { aList index } {
#		Answer the integer value at the element-th entry of
#		a list and the index-th entry of that element.

	binary scan [lindex $aList $index] i result
	if { $result < 0} then {
		set result [expr 2147483648 + $result]
	}
	return $result
}

proc Arena2Texture::open { tturePath } {
#		Open the requested texture pathname and return its handle.

	variable GameDir

	set path [file join "$GameDir" "$tturePath"]
	if {[file exists $path]} then {
		set fid [::open $path r]
		fconfigure $fid -translation binary
		return $fid
	}
	return -1
}

proc Arena2Texture::readHeader { ttureFID } {
#		Read the next texture header from the present position
#		in the file.

	return [read $ttureFID 20]
}

proc Arena2Texture::reportHeader { header } {
#		Document the contents of a texture header.

	variable headerType1
	variable headerType2
	variable headerOffset
	variable headerMystery

#	type1 seems to be a palette index number.
	set type1 [shortAt $header $headerType1]
#	type2 seems to point to the last address of
#	a multiplex image on a canvas.
	set type2 [longAt $header $headerType2]
	set off [longAt $header $headerOffset]
#	unk is uniformly 0.
#	set unk [shortAt $header $headerMystery]
	puts stdout [format "type: 0x%04x 0x%06x, offset: 0x%08x" $type1 $type2 $off]
	return 1
}

proc Arena2Texture::reportHeaders { headers serialNo } {
#		Report the contents of the current texture file headers.

	set index 0
	set length [llength $headers]
	while {$index < $length} {
		set header [lindex $headers $index]
		puts -nonewline stdout "-h [format %s:%-3d $serialNo $index] "
		reportHeader "$header"
		incr index
	}
	return 1
}

proc Arena2Texture::reportImageHeader { header } {
#		Document the contents of a texture image header.

	variable xOffset
	variable yOffset
	variable width
	variable height
	variable imageOffset
	variable imageSize
	variable subimages
	variable headerStart
	variable compression
	variable interlace
	variable frameDelay
	variable imageScale

	if { ! [llength $header]} then {
		return 0
	}
	set xOff [shortAt $header $xOffset]
	if {$xOff > 32768} then {
		set xOff [expr $xOff - 65536]
	}
	set yOff [shortAt $header $yOffset]
	if {$yOff > 32768} then {
		set yOff [expr $yOff - 65536]
	}
	set w [shortAt $header $width]
	set h [shortAt $header $height]
	set off [longAt $header $imageOffset]
	set size [longAt $header $imageSize]
	set subs [shortAt $header $subimages]
	set start [lindex $header $headerStart]
	set soff [expr $start + $off]
	set cpr ""
	set u2 [shortAt $header $interlace]
	set u3 [shortAt $header $frameDelay]
#	u4 are the x and y scale factors for the image.
	set u4 [longAt $header $imageScale]
	set u4x [expr $u4 & 0x7fff]
	set u4y [expr $u4 >> 16]
	set compress [shortAt $header $compression]

	if { $subs > 1 } then {
		set cpr " flipbook  "
	} elseif { $compress == 264 || $compress == 4360} then {
		set cpr " row-cmprs "
	} elseif { $compress == 2304 || $compress == 256 || $compress == 257} then {
		set cpr " canvas    "
	} elseif { $compress == 0 && $u2 > 0} then {
		set cpr " canvas    "
	} else {
		set cpr [format " 0x%04x    " $compress]}
	puts stdout [format "(%3d,%3d) %3dx%3d %5d%s 0x%04x 0x%04x 0x%04x 0x%04x, %2d images at %#x" $xOff $yOff $w $h $size $cpr $u2 $u3 $u4x $u4y $subs $soff]
	return 1
}

proc Arena2Texture::reportImageHeaders { headers serialNo } {
#		Report the contents of the current texture file image headers.

	set index 0
	set length [llength $headers]
	while {$index < $length} {
		set header [lindex $headers $index]
		puts -nonewline stdout "-i [format %s:%-3d $serialNo $index] "
		reportImageHeader "$header"
		incr index
	}
	return 1
}

proc Arena2Texture::shortAt { aList index } {
#		Answer the integer value at the element-th entry of
#		a list and the index-th entry of that element.

	binary scan [lindex $aList $index] s result
	if { $result < 0 } then {
		set result [expr 65536 + $result]
	}
	return $result
}

proc Arena2Texture::type { ttureFID } {
#		Answer the type code for this texture file.

	seek $ttureFID 0 start
	set type [read $ttureFID 2]
	return "$type"
}

#	Inherit the installation configuration from the Arena2Config
#	namespace when it is available.

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Config::artDir
	variable Arena2Config::exeDir
	variable Arena2Config::tmpDir
	variable Arena2Texture::GameDir "$Arena2Config::artDir"
	variable Arena2Texture::defaultRate "$Arena2Config::frameRate"
	variable Arena2Texture::picFile [file join "$Arena2Config::tmpDir" "pic"]
	variable Arena2Texture::ppmFile [file join "$Arena2Config::tmpDir" "pic.00" ]
}
#Arena2Texture::browseTable 432
eval $argv
