/*
						 ***********
						** Summary **
						 ***********
	(1)	Title:
				Texture to PPM Converter

	(2)	Purpose:
			To decompress packed imaged from texture.???
			and *.cif files in Daggerfall.

			There are 6 fundamental conversions supported:

				o  palette to true color
				o  indexed RLE (for *.cif animations)
				o  whole image RLE
				o  base image and deltas
				o  row by row RLE
				o  interlaced canvas (scrapbook or picture album)

			The compression type field of the textures indicate
			a wider variety of forms, but for conversion to individual
			pixmaps, these are sufficient.

	(3)	Non-ANSI features:
*/
#	pragma pack(1)
/*
	(4)	Platform dependencies:
			Win32 can't cope at all with binary stdin.  So the filter
			has sprouted a requirement for an input file name containing
			the extracted image data..

	(5)	Change history:
			***  None  ***

	(6)	Input:

			Structures:
				***  None  ***

			Common:
				***  None  ***

			Files:
				***  None  ***

			Assertions about input:
				Command line syntax:

					txttoppm  -p aPaletteFile  \
					          width  height  chunkSize \
					          compression  frames  \
					          filePattern  sourceTxt

				where the -p switch specifies the 256 color
				palette to use to convert the encoded pixels to true
				color.

				width is the true width of the image in pixels.

				height is the true height of the image in pixels.

				chunkSize is the number of bytes to process from
				stdin.  For many compression schemes, chunkSize
				is not width * height.

				compression specifies the type of image compression
				used to create the stdin data stream.

				frames specifies the number of frames when an
				animation flipbook is being decompressed.

				filePattern specifies the output file pattern to use
				for the decompressed pixmaps.  It must have the
				form

					aPath.00

				Each image in a flip book will be written to successive
				files based on the pattern.  So a 3 image flip book
				will be stored as

					aPath.00
					aPath.01
					aPath.02

				which is the form expected by the Tk/Tcl animator.

				sourecTxt specifies the temporary file containing the
				image data to decompress.

	(7)	Output:

			Structures:
				***  None  ***

			Common:
				***  None  ***

			Files:
				***  None  ***

			Assertions about the output:
				Tk/Tcl only understands the `raw' flavor of ppm pixmaps.

	(8)	Public members:

			Title:

			Purpose:

			Parameters:

	(9)	Private members:

			Title:

			Purpose:

			Parameters:

	(10) Subprograms called:
			***  None  ***

	(11) Includes referenced:
			***  None  ***

	(12) Structures declared:
			***  None  ***

						 ****************
						** Declarations **
						 ****************

					 ***********************
					** Global declarations **
					 ***********************

				  Copyright 2001, Nagual Enterprises
			Please see COPYING for the complete GNU license

					  ********************
					 ** Global functions **
					  ********************
*/

#					include <stdlib.h>
#					include <stdio.h>
#					include <sys/types.h>
#					include <sys/stat.h>
#					include <fcntl.h>
#					include <string.h>

/*
					  *************
					 ** Constants **
					  *************
*/

#				ifndef O_BINARY
#					define O_BINARY 0
#					define setmode(a,b)
#				endif

/*
					   **********
					  ** Macros **
					   **********
*/

#					define bail(x)	{fprintf(stderr, x); exit(1);}
#					define warn(x)	fprintf(stderr, x)

/*
					 **************
					** Structures **
					 **************
*/

				struct _IMGHeader
				{
					unsigned short canvasX;		/*	Canvas width? */
					unsigned short canvasY;		/*	Canvas height? */
					unsigned short imageWidth;
					unsigned short imageHeight;
					unsigned short zip1;		/*	Always 0; but we usurp its
													meaning internally to tag
													compressed images. */
					unsigned long imageSize;	/*	Count of subsequent bytes,
													which may be compressed. */
				};
				typedef struct _IMGHeader IMGHeader;

/*
					   **********
					  ** Common **
					   **********
*/

					IMGHeader header;
					unsigned char ArenaPalette[256][3];
					unsigned char* image = NULL;
					long outputSize = 0;
					int subimages = 1;
					FILE* Stdin = NULL;

/*
				 **********************
				** Local declarations **
				 **********************

					  *************
					 ** Variables **
					  *************

					***  None  ***

						 ********
						** Text **
						 ********


							 *******************
							** Private methods **
							 *******************
*/

/*
							 ******************
							** Public methods **
							 ******************
*/

int buildPalette
(	char* path)
{
		/*	Build the palette index from the contents of a path.  Answer
			false if the palette could not be constructed, and true
			otherwise. */

	int pal;
	struct stat pStat;
	int size;

	if (stat(path, &pStat) < 0)
	{
		/*	Gripe if the requested file cannot be located. */

		perror("txttoppm");
		return 0;
	}
	/*	Confirm this smells like a palette file. */

	size = pStat.st_size;
	if (size != 768 && size != 776)
	{
		fprintf(stderr, "Palette file %s is not 768 or 776 bytes.\n", path);
		return 0;
	}
	pal = open(path, O_RDONLY | O_BINARY);
	if (!pal)
	{
		fprintf(stderr, "Can't open palette file %s.\n", path);
		return 0;
	}
	/*	Skip the preamble on palette files which have one. */

	if (size == 776)
		lseek(pal, 8, SEEK_SET);
	if (768 != read(pal, ArenaPalette, 768))
	{
		fprintf(stderr, "Palette file %s is not 768 bytes.\n", path);
		close(pal);
		return 0;
	}
	close(pal);
	return 1;
}

unsigned char* decompressChunk
(	unsigned char* newData,
	unsigned char* data,
	long rawSize)
{
		/*	Stock RLE decompression of rawSize bytes from data
			into newData.  Answer the next available position
			in newData. */

	while (rawSize > 0)
	{
		if (data[1] > 127)
		{
			short count = -*(short*)data;
			data = &data[2];
			rawSize -= 3;
			while (count--)
				*newData++ = *data;
			data++;
		}
		else
		{
			short count = *(short*)data;
			if (count > header.imageWidth)
				break;
			rawSize -= count + 2;
			data = &data[2];
			while (count--)
				*newData++ = *data++;
		}
	}
	return newData;
}

unsigned char* decompressRLE
(	unsigned char* newData,
	unsigned char* data,
	long rawSize)
{
		/*	Stock RLE decompression of rawSize bytes from data
			into newData.  Answer the next available position
			in newData. */

	while (rawSize > 0)
	{
		if (*data > 127)
		{
			short count = *data - 127;
			data = &data[1];
			rawSize -= 2;
			while (count--)
				*newData++ = *data;
			data++;
		}
		else
		{
			short count = 1 + (short)*(char*)data;
			if (count > header.imageWidth)
				break;
			rawSize -= count + 1;
			data = &data[1];
			while (count--)
				*newData++ = *data++;
		}
	}
	return newData;
}

unsigned char* decompressFrames
(	unsigned char* newData,
	unsigned char* data,
	long rawSize)
{
		/*	Decompress a series of RLE commpressed franes,
			indexed by relative offset. */

	int images = subimages;
	unsigned short* head = (unsigned short*)&data[12];
	unsigned short* last = (unsigned short*)&data[12+62];

	while (images--)
	{
		unsigned char* image = &data[*head];
		long size;
		size = images ?
			(long)(head[1] - *head) :
			*last - *head;
		newData = decompressRLE(newData, image, size);
		head++;
	}
	return newData;
}

unsigned char* decompressByRow
(	unsigned char* newData,
	unsigned char* data,
	unsigned long rawSize)
{
		/*	Decompress an image file stored with row-by-row compression.
			Note that the algorithm allows rows that compress to identical
			bit streams to be shared amongst subimages. */

#ifdef __BORLANDC__
	short hashedLength[512];
#else
	short hashedLength[rawSize+2];
#endif
	unsigned char* fouData = &data[-28];
	unsigned char* rowData;
	unsigned short* head = (unsigned short*)data;
	int images = subimages;
	unsigned short offset = 0;
	unsigned long imageBytes = rawSize;

	memset(hashedLength, 0, sizeof(hashedLength));
	while (images--)
	{
		int count = header.imageHeight;
		while (count--)
		{
			int raw;
			int size = 0;
			int update = 1;
			offset = *head++ + 2;
			raw = *head++ == 0x8000;
			if (hashedLength[offset])
			{
				size = hashedLength[offset];
				update = 0;
			}
			else
				size = (int)*head - (int)offset;
			if (size < 0)
			{
				int rowLeft = count + (images * header.imageHeight);
				unsigned short* lastHead = &head[2];
				if (images)
					rowLeft++;
				else if (!count)
					update = 1;
				if (rowLeft)
				{
					while (rowLeft--)
					{
						size = (int)*lastHead - (int)offset;
						if (size >= 0)
							break;
						lastHead = &lastHead[2];
					}
					if (size < 0)
						size = imageBytes - offset;
					update = 0;
				}
			}
			hashedLength[offset] = size;
			rowData = &fouData[offset];

			/*	'34' is a bit mysterious below.  == 6 + 28.  28, for the
				size of the image head record, 6 for the regular offset
				magic. */

			if (!count && 1 == subimages)
				size = rawSize - 34;
				/*size = imageBytes - offset;*/
			if (update)
				rawSize -= size + 6;
			if (raw)
			{
				/*	This row is RLE compressed.  Retrieve its plain stream
					of byte codes. */

				newData = decompressChunk(newData, rowData, size);
			}
			else
			{
				/*	Uncompressed rows don't have a row size field, so
					compensate for the extra space in the computed size. */

				size = header.imageWidth;
				rowData = &rowData[-2];
				while (size--)
					*newData++ = *rowData++;
			}
		}
	}
	/*	Note, rawSize must == 28 at loop end, else a decoding error
		has occurred.  Subimages (animations) follow different semantics
		though. */

	return newData;
}

unsigned char* deinterlace
(	unsigned char* newData,
	unsigned char* data)
{
		/*	Demultiplex an image stored on a common 256 x whatever
			canvas.  The byte stream is uncompressed. */

	int count;
	int skip;
	int rows = header.imageHeight;

	skip = 256 - header.imageWidth;
	while (rows--)
	{
		count = header.imageWidth;
		while (count--)
			*newData++ = *data++;
		data = &data[skip];
	}
	return newData;
}

unsigned char* decompressByImage
(	unsigned char* newData,
	unsigned char* data)
{
		/*	Decompress an image built from several subimages or parts.
			(Generally, such textures will animate as lamp flames,
			etc) */

	int count = subimages;
	long* offsets = (long*)data;

	while (count--)
	{
		short w = *(short*)&data[*offsets];
		short h = *(short*)&data[*offsets+2];
		unsigned char* row = &data[*offsets+4];
		int rows = h;
		int repeats;
		while (rows--)
		{
			int columns = w;
			do
			{
				repeats = *row++;
				columns -= repeats;
				while (repeats--)
					*newData++ = '\0';
				if (columns > 0)
				{
					repeats = *row++;
					columns -= repeats;
					while (repeats--)
						*newData++ = *row++;
				}
				else
					row++;
			} while (columns > 0);
		}
		offsets++;
	}
	return newData;
}

int decompress
()
{
		/*	Decrypt a compressed image format.  Answer true if the
			decryption has succeeded, and false otherwise.  Replace
			the image contents upon success, discarding the original. */

	unsigned char* newImage;
	long imageType = header.zip1;
	outputSize = header.imageWidth * header.imageHeight * subimages;
	newImage = malloc(outputSize);
	if (subimages > 1)
		imageType |= 0x10000;
	switch (imageType)
	{
		case 0:
		case 1:
		case 0x0800:
			/*	Apply just the palette transform. */

			free(newImage);
			return 1;
			break;

		case 0x10002:
			/*	Synthetic compression type to capture CIF animations.
				The actual compression type, found in exactly 1 of
				the Cif animations, appears to be 3. */

			decompressFrames(newImage, image, header.imageSize);
			break;

		case 0x0002:
			/*	Synthetic compression type to capture CIF compression. */

			decompressRLE(newImage, image, header.imageSize);
			break;

		case 0x10900:
			/*	Decompress a texture image stored as a series of
				of image sections (deltas) and subimages. */

			decompressByImage(newImage, image);
			break;

		case 0x10108:
		case 0x1108:
		case 0x0108:
			/*	Decompress a texture image stored with row by row
				RLE compression. */

			decompressByRow(newImage, image, header.imageSize);
			break;

		case 0x0900:
		case 0x0101:
		case 0x0100:
			/*	Deconvolute an uncompressed image multiplexed with other
				pictures on a 256 x whatever canvas.  Texture files are
				generally written so that the multiplexed canvas images
				are recorded firstmost in the texture file, preceeding
				any RLE compressed images. */

			deinterlace(newImage, image);
			break;

		default:
			fprintf(stderr, "Unsupported compression format %#x.\n", header.zip1);
			return 0;
	}
	free(image);
	image = newImage;
	return 1;
}

int readImg
(	int needHdr)
{
		/*	Retrieve the stream of image bytes from stdin.  If needHdr
			is true, the byte stream contains the image header.
			Otherwise, the image information is already prefabricated
			in the header data structure.  Answer true on successful
			image retrieval, and false otherwise. */

	if (needHdr)
	{
		header.imageSize = 0L;
		if (fread((void*)&header, sizeof(header)-2, 1, Stdin) != 1)
		{
			perror("img2ppm");
			fprintf(stderr, "Unable to read the image header.\n");
			exit(1);
		}
		if (header.imageHeight == 0 || header.imageWidth == 0)
			bail("IMG cannot have image height or image width of 0.\n")
		if (header.imageSize != (header.imageHeight * header.imageWidth))
		{
			warn("IMG file size isn't a multiple of the image height and width.\n");
			fprintf(stderr, "(%d, %d) != %d.\n",
					header.imageWidth,
					header.imageHeight,
					header.imageSize);
		}
	}
	image = malloc(header.imageSize);
	outputSize = header.imageSize;
	fread(image, header.imageSize, 1, Stdin);
	/*if (fread(image, header.imageSize, 1, Stdin) != 1)
		bail("Not enough data for IMG.\n")*/
	return header.zip1 ?
		decompress() :
		1;
}

int main
(	int argc,
	char** argv)
{
		/*	Transform an Arena image on stdin to PPM on stdout.  Answer
			0 on successful transformation, and != 0 on failure. */

	int needHdr = 1;
	FILE* out = stdout;
	char* base;
	char* end;
	unsigned char* bits;
	char* palFile = "/dos/g/game/dag/arena2/art_pal.col";
	/*char* palFile = "/dos/g/game/dag/arena2/pal.pal";*/
	/*char* palFile = "/dos/g/game/dag/arena2/pal.raw";*/
	int count;
	int index = 0;

	argv++;
	argc--;
	if (argc > 0 && argv[0][0] == '-' && argv[0][1] == 'p')
	{
		/*	Obtain the palette file override from the cli. */

		argv++;
		argc--;
		palFile = *argv;
		argv++;
		argc--;
	}
	if (argc > 4)
	{
		/*	For texture-based images, pass
				width  height  size  compress  subimages
			on the command line.  'compress' is the compression
			style from the texture image header. */

		header.imageWidth = strtol(*argv++, NULL, 0);
		header.imageHeight = strtol(*argv++, NULL, 0);
		header.imageSize = strtol(*argv++, NULL, 0);
		header.zip1 = strtol(*argv++, NULL, 0);
		needHdr = header.zip1 == 1;
		subimages = strtol(*argv++, NULL, 0);

		/*	Grab the destination file pattern.  For animations, the
			trailing `extension' should be .00 or .000.  The name
			will be `autoincremented' to cover the separate images in
			the animation flipbook. */

		base = *argv++;
		end = &base[strlen(base) - 2];

		/*	Grab the source file name.  Win32 can't pipe stdin with
			any pretense of reliablility, much less cope with binary
			data. (the open file handle across WinProc bug) */

		Stdin = fopen(*argv, "rb");
		if (!Stdin)
		{
			perror("img2ppm");
			fprintf(stderr, "Can't open %s for input.\n", *argv);
			exit(1);
		}
		if (subimages < 0)
		{
			fseek(Stdin, -subimages, SEEK_SET);
			subimages = 1;
		}
	}
	/*	Install the specified true-color mapping and slurp in the
		whole image stream. */

	if (!buildPalette(palFile))
		exit(1);
	if (!readImg(needHdr))
		return 1;

	/*	Burst the resulting pixmap(s) into separate files for
		the animator.  Note that Tk only groks raw PPM formats. */

	count = subimages == 0 ?
		1 :
		subimages;
	bits = image;
	outputSize /= count;
	while (count--)
	{
		div_t ext;
		int bytes = outputSize;

		out = fopen(base, "wb");
		if (!out)
		{
			perror("img2ppm");
			fprintf(stderr, "Can't open %s for output.\n", base);
			exit(1);
		}
		fprintf(out, "P6\n%d %d\n%d\n", header.imageWidth, header.imageHeight, 255);
		while (bytes--)
		{
			unsigned char* color = ArenaPalette[*bits++];
			fwrite(color, 3, 1, out);
		}
		fclose(out);
		index++;
		ext = div(index, 10);
		end[1] = ext.rem + 48;
		*end = ext.quot + 48;
	}
	return 0;
}

